#include "nlp_socket.h"
#include <nlp_macros.h>
#include <assert.h>
#include <errno.h>


Socket_set_t AllSockets::set_AllSockets;


//															
void AllSockets::Add (Socket* _pSocket)
{
	assert (set_AllSockets.end () == set_AllSockets.find (_pSocket));
	set_AllSockets.insert (_pSocket);
}



//															
void AllSockets::Remove (Socket* _pSocket)
{
	assert (set_AllSockets.end () != set_AllSockets.find (_pSocket));
	set_AllSockets.erase (_pSocket);
}



//															
void AllSockets::CloseAndDeleteAllSockets (void)
{
	ITERATE (Socket_set_t, set_AllSockets, ite)
		(*ite)->Close ();
	set_AllSockets.clear ();
}



//															
void AllSockets::ProcessEvents (unsigned long _lMillisecTimeout)
{
	fd_set	fsReadReady;
	fd_set	fsWriteReady;
	fd_set	fsError;

	FD_ZERO (&fsReadReady);
	FD_ZERO (&fsWriteReady);
	FD_ZERO (&fsError);


	int iMaxFD = -1;
	ITERATE (Socket_set_t, set_AllSockets, ite)
	{
		Socket* pSocket = *ite;
		// standalone sockets are meant for multi-threading	
		// and have to do their own even processing.		
		if (true == pSocket->b_Standalone)
			continue;

		if ((SOCKET_FD_UNDEF == pSocket->fd_Socket) ||
			(SOCKET_FD_CLOSED == pSocket->fd_Socket) ||
			(SOCKET_FD_DESTROYED == pSocket->fd_Socket))
			continue;

		FD_SET (pSocket->fd_Socket, &fsReadReady);
		FD_SET (pSocket->fd_Socket, &fsError);

		// check for write ready only for sockets waiting	
		// to send data ...									
		if (pSocket->o_SendBuffer.Length () > 0)
			FD_SET (pSocket->fd_Socket, &fsWriteReady);

		// checking for write ready on non-blocking sockets	
		// in the process of being connected tell us when	
		// they do get connected ...						
		if (Socket::InProgress == pSocket->e_ConnectionStatus)
			FD_SET (pSocket->fd_Socket, &fsWriteReady);

		if (iMaxFD < pSocket->fd_Socket)
			iMaxFD = pSocket->fd_Socket;
	}

	timeval oTimeout;
	oTimeout.tv_sec = _lMillisecTimeout / 1000;
	oTimeout.tv_usec = (_lMillisecTimeout * 1000) % 1000000;

	select (iMaxFD + 1, &fsReadReady, &fsWriteReady, &fsError, &oTimeout);

	ITERATE (Socket_set_t, set_AllSockets, ite)
	{
		Socket* pSocket = *ite;
		// standalone sockets are meant for multi-threading	
		// and have to do their own even processing.		
		if (true == pSocket->b_Standalone)
			continue;

		// Note that a socket might become disconnected through	
		// the functionality of inherited classes in response	
		// to the virtual method calls made here. So we need	
		// to check after each call if the socket is still		
		// connected. 											


		// check for sockets that are read-ready ...			
		if ((SOCKET_FD_UNDEF == pSocket->fd_Socket) ||
			(SOCKET_FD_CLOSED == pSocket->fd_Socket) ||
			(SOCKET_FD_DESTROYED == pSocket->fd_Socket))
			continue;

		if (true == FD_ISSET (pSocket->fd_Socket, &fsReadReady))
		{
			// OnData is a virtual method overloaded by	
			// ServerSocket to accept incomming socket	
			// connections ...							
			pSocket->OnData ();
		}


		// check for sockets that are write-ready ...			
		if ((SOCKET_FD_UNDEF == pSocket->fd_Socket) ||
			(SOCKET_FD_CLOSED == pSocket->fd_Socket) ||
			(SOCKET_FD_DESTROYED == pSocket->fd_Socket))
			continue;

		if (true == FD_ISSET (pSocket->fd_Socket, &fsWriteReady))
		{
			if (Socket::Connected == pSocket->e_ConnectionStatus)
				pSocket->OnCanSend ();

			else if (Socket::InProgress == pSocket->e_ConnectionStatus)
			{
				int iOptionValue = -1;
				socklen_t iOptionLength = sizeof (iOptionLength);
				if (0 != getsockopt (pSocket->fd_Socket, SOL_SOCKET, SO_ERROR, 
									 &iOptionValue, &iOptionLength))
				{
					String sError (strerror (errno));
					cerr << "[ERROR]  getsockopt returned error when checking "
						    "socket connection in progress."
						 << endl;
				}
				if (0 == iOptionValue)
				{
					pSocket->e_ConnectionStatus = Socket::Connected;
					pSocket->OnConnect ();
				}
			}
			else
			{
				cerr << "[ERROR]  Received write-ready notification on socket "
						"that is not connected."
					 << endl;
				abort ();
			}
		}


		// check for sockets that are in error ...				
		if ((SOCKET_FD_UNDEF == pSocket->fd_Socket) ||
			(SOCKET_FD_CLOSED == pSocket->fd_Socket) ||
			(SOCKET_FD_DESTROYED == pSocket->fd_Socket))
			continue;

		if (true == FD_ISSET (pSocket->fd_Socket, &fsError))
			pSocket->OnError ();
	}

	FD_ZERO (&fsReadReady);
	FD_ZERO (&fsWriteReady);
	FD_ZERO (&fsError);
}



